#pragma once

#include "configuration.hh"
#include "ios_precomputers/delegate.hh"
#include "ios_precomputers/fake_vars.hh"
#include "ios_precomputers/powset.hh"
#include "ios_precomputers/standard.hh"
#include "ios_precomputers/mona.hh"
